Imports System.ComponentModel
Imports System.Runtime.CompilerServices
Imports System.Threading
Imports System.Threading.Tasks

Public Class AsyncCommand(Of TResult)
	Inherits AsyncCommandBase
	Implements INotifyPropertyChanged
	Private ReadOnly _command As Func(Of CancellationToken, Task(Of TResult))
	Private ReadOnly _cancelCommand As CancelAsyncCommand
	Private _execution As NotifyTaskCompletion(Of TResult)

	Public Sub New(ByVal command As Func(Of CancellationToken, Task(Of TResult)))
		_command = command
		_cancelCommand = New CancelAsyncCommand()
	End Sub

	Public Overrides Function CanExecute(ByVal parameter As Object) As Boolean
		Return Execution Is Nothing OrElse Execution.IsCompleted
	End Function

	Public Overrides async Function ExecuteAsync(ByVal parameter As Object) As Task
		_cancelCommand.NotifyCommandStarting()
		Execution = New NotifyTaskCompletion(Of TResult)(_command(_cancelCommand.Token))
		RaiseCanExecuteChanged()
		await Execution.TaskCompletion
		_cancelCommand.NotifyCommandFinished()
		RaiseCanExecuteChanged()
	End Function

	Public ReadOnly Property CancelCommand() As ICommand
		Get
			Return _cancelCommand
		End Get
	End Property

	Public Property Execution() As NotifyTaskCompletion(Of TResult)
		Get
			Return _execution
		End Get
		Private Set(ByVal value As NotifyTaskCompletion(Of TResult))
			_execution = value
			OnPropertyChanged()
		End Set
	End Property

    Public Event PropertyChanged As PropertyChangedEventHandler Implements INotifyPropertyChanged.PropertyChanged
    Protected Overridable Sub OnPropertyChanged(<CallerMemberName> Optional ByVal propertyName As String = Nothing)
        Dim handler As PropertyChangedEventHandler = PropertyChangedEvent
        If handler IsNot Nothing Then

            handler(Me, New PropertyChangedEventArgs(propertyName))
        End If
    End Sub

	Private NotInheritable Class CancelAsyncCommand
		Implements ICommand
		Private _cts As New CancellationTokenSource()
		Private _commandExecuting As Boolean

		Public ReadOnly Property Token() As CancellationToken
			Get
				Return _cts.Token
			End Get
		End Property

		Public Sub NotifyCommandStarting()
			_commandExecuting = True
			If Not _cts.IsCancellationRequested Then
				Return
			End If
			_cts = New CancellationTokenSource()
			RaiseCanExecuteChanged()
		End Sub

		Public Sub NotifyCommandFinished()
			_commandExecuting = False
			RaiseCanExecuteChanged()
		End Sub

		Private Function CanExecute(ByVal parameter As Object) As Boolean Implements ICommand.CanExecute
			Return _commandExecuting AndAlso Not _cts.IsCancellationRequested
		End Function

		Private Sub Execute(ByVal parameter As Object) Implements ICommand.Execute
			_cts.Cancel()
			RaiseCanExecuteChanged()
		End Sub

		Public Custom Event CanExecuteChanged As EventHandler Implements ICommand.CanExecuteChanged
			AddHandler(ByVal value As EventHandler)
				AddHandler CommandManager.RequerySuggested, value
			End AddHandler
			RemoveHandler(ByVal value As EventHandler)
				RemoveHandler CommandManager.RequerySuggested, value
			End RemoveHandler
			RaiseEvent(ByVal sender As Object, ByVal e As EventArgs)
                CanExecute(sender)
			End RaiseEvent
		End Event

		Private Sub RaiseCanExecuteChanged()
			CommandManager.InvalidateRequerySuggested()
		End Sub
	End Class
End Class

Public NotInheritable Class AsyncCommand
	Private Sub New()
	End Sub
	Public Shared Function Create(ByVal command As Func(Of Task)) As AsyncCommand(Of Object)
        Return New AsyncCommand(Of Object)(Async Function()
                                               Await command()
                                               Return Nothing
                                           End Function)
	End Function

	Public Shared Function Create(Of TResult)(ByVal command As Func(Of Task(Of TResult))) As AsyncCommand(Of TResult)
        Return New AsyncCommand(Of TResult)(Function() command())
	End Function

	Public Shared Function Create(ByVal command As Func(Of CancellationToken, Task)) As AsyncCommand(Of Object)
        Return New AsyncCommand(Of Object)(Async Function(token)
                                               Await command(token)
                                               Return Nothing
                                           End Function)
	End Function

	Public Shared Function Create(Of TResult)(ByVal command As Func(Of CancellationToken, Task(Of TResult))) As AsyncCommand(Of TResult)
		Return New AsyncCommand(Of TResult)(command)
	End Function
End Class
